import pytest
from nanomotif.binnary import data_processing
from .conftest import MockArgs
import polars as pl

def test_feature_with_loaded_data(loaded_data):
    # Access the loaded data directly if returned as a dictionary
    motifs_scored = loaded_data["motifs_scored"]
    bin_motifs = loaded_data["bin_motifs"]
    contig_bins = loaded_data["contig_bins"]

    # Now you can use the data in your test assertions
    assert motifs_scored is not None

    # bin motifs
    assert bin_motifs is not None
    assert set(bin_motifs["bin"].unique()) == {"b1", "b2", "b3"}

    # contig bins
    assert contig_bins is not None
    assert set(contig_bins["bin"].unique()) == {"b1", "b2", "b3", "b4"}

    contig_set = {f"contig_{i}" for i in range(1, 14)} | {"contig_16"}
    assert set(contig_bins["contig"].unique()) == contig_set



def test_prepare_bin_consensus(loaded_data):
    """
    GIVEN loaded_data
    WHEN prepare_bin_consensus is called
    THEN assert that the output contains only the expected columns
    """
    args = MockArgs()
    
    bin_motif_binary = data_processing.prepare_bin_consensus(loaded_data["bin_motifs"], args)
    
    assert bin_motif_binary is not None
    assert bin_motif_binary.columns == ["bin", "motif_mod", "mean_methylation", "methylation_binary"]
    
    assert bin_motif_binary.filter((pl.col("bin") == "b3") & (pl.col("motif_mod") == "m6_a-1")).select("methylation_binary").item() == 1
    
    # Assert that there are 4 motifs in bin 1
    assert bin_motif_binary.filter(pl.col("bin") == "b3").height == 4
    assert set(bin_motif_binary.filter(pl.col("bin") == "b3").to_pandas()["motif_mod"].to_list()) == {"m1_a-1", "m2_a-1", "m3_a-1", "m6_a-1"}
    
    # Assert m7_a is filtered because of too few observations
    assert bin_motif_binary.filter(pl.col("motif_mod") == "m7_a-1").is_empty()
    


def test_motifs_scored_in_bins(loaded_data):
    # Access the loaded data directly if returned as a dictionary
    motifs_scored = loaded_data["motifs_scored"]
    bin_motifs = loaded_data["bin_motifs"]
    contig_bins = loaded_data["contig_bins"]

    # Step 1 create bin_motif_binary
    args = MockArgs()
    bin_motif_binary = data_processing.prepare_bin_consensus(bin_motifs, args)

    # Step 2: create motifs_scored_in_bins
    motifs_scored_in_bins = data_processing.prepare_motifs_scored_in_bins(
        motifs_scored,
        bin_motif_binary.select("motif_mod").unique()["motif_mod"],
        contig_bins
    )

    # motifs_scored_in_bins
    assert sorted(bin_motif_binary.select("motif_mod").unique()["motif_mod"]) == sorted(
        motifs_scored_in_bins.select("motif_mod").unique()["motif_mod"]
    )
    
    # Assert that the number of columns is 12
    assert len(motifs_scored_in_bins.columns) == 12  # number of columns
    # Assert contig 1 belongs to bin 1
    assert motifs_scored_in_bins.filter(pl.col("contig") == "contig_1").select("bin").unique()["bin"].item() == "b1"
    



def test_remove_ambiguous_motifs(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data and motifs_scored_in_bins_and_bin_motifs
    WHEN construct_bin_motifs_from_motifs_scored_in_bins is called
    THEN assert that the output contains only the expected columns
    """
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    args = MockArgs()
    
    
    bins_w_no_ambiguous_motifs = data_processing.remove_ambiguous_motifs_from_bin_consensus(
        motifs_scored_in_bins,
        args
    )
    print(bins_w_no_ambiguous_motifs)
    
    print(bins_w_no_ambiguous_motifs.filter(pl.col("bin") == "b1").to_pandas())
    assert bins_w_no_ambiguous_motifs is not None
    assert bins_w_no_ambiguous_motifs.columns == ['bin', 'motif_mod']
    assert bins_w_no_ambiguous_motifs.filter((pl.col("bin") == "b1") & (pl.col("motif_mod") == "m1_a")).is_empty()
    
    


def test_bin_motifs_from_motifs_scored_in_bins(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data and motifs_scored_in_bins_and_bin_motifs
    WHEN construct_bin_motifs_from_motifs_scored_in_bins is called
    THEN assert that the output contains only the expected columns
    """
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    args = MockArgs()
    print(motifs_scored_in_bins)
    bin_motifs_from_motifs_scored_in_bins = data_processing.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    assert bin_motifs_from_motifs_scored_in_bins is not None
    assert bin_motifs_from_motifs_scored_in_bins.columns == ['bin', 'motif_mod', 'n_mod', 'n_nomod', 'n_motifs_bin', 'mean_methylation', 'mean_methylation_bin', 'std_methylation_bin', 'n_contigs', 'methylation_binary']
    
    
# def test_calculate_binary_motif_comparison_matrix(loaded_data, motifs_scored_in_bins_and_bin_motifs):
#     """
#     GIVEN loaded_data and motifs_scored_in_bins_and_bin_motifs
#     WHEN calculate_binary_motif_comparison_matrix is called
#     THEN assert that the output contains only the expected columns
#     """
#     motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
#     args = MockArgs()
    
#     motifs_scored_in_bins_filtered = motifs_scored_in_bins \
#         .filter(~pl.col("bin_contig").str.contains("unbinned"))
    
#     motif_binary_compare = data_processing.calculate_binary_motif_comparison_matrix(
#         motifs_scored_in_bins_filtered,
#         args
#     )
    
#     assert motif_binary_compare is not None
#     # Assert that no bin_contig contains "unbinned"
#     assert motif_binary_compare.filter(pl.col("bin_compare").str.contains("unbinned")).is_empty()
    
#     # b3 = motif_binary_compare[(motif_binary_compare["bin"] == "b3") & (motif_binary_compare["bin_compare"].str.contains("b3"))]
#     b3 = motif_binary_compare.filter((pl.col("bin") == "b3") & (pl.col("bin_compare").str.contains("b3")))
    
#     assert set(b3.to_pandas()["motif_mod"].unique()) == set(["m1_a", "m2_a", "m3_a", "m6_a"])
#     assert b3.filter(pl.col("motif_mod") == "m6_a").select("methylation_binary").get_column("methylation_binary")[0] == 1
#     assert b3.filter(pl.col("motif_mod") == "m2_a").select("methylation_binary").get_column("methylation_binary")[0] == 0
    


import polars as pl
import re

def split_bin_contig(df):
    df = df \
        .with_columns([
            pl.col("bin_compare").str.extract(r"(.+)_contig_\d+", 1).alias("contig_bin"),
            pl.col("bin_compare").str.extract(r"contig_(\d+)", 1).alias("contig_number")
        ]) \
        .with_columns(
            ("contig_" + pl.col("contig_number")).alias("contig")
        ) \
        .drop("contig_number")

    return df


def get_bin(bin_contig):
    return re.search(r"(.+)_contig_\d+", bin_contig).group(1)


# pyright: reportUndefinedVariable=false
#%% Load modules
import pandas as pd
from Bio import SeqIO

#%% Importing defenseFinder hmmer and systems output
df_hmmer_path = snakemake.input['hmmer_file']
df_hmmer = pd.read_csv(df_hmmer_path, sep='\t')

df_systems_path = snakemake.input["system_file"]
df_systems = pd.read_csv(df_systems_path, sep='\t')

#%% Filter defensefinder data for RM systems only.
df_hmmer_RM = df_hmmer[df_hmmer['gene_name'].str.contains('_Type_', na=False)]

# Group by gene_id (hit id), then retrieve the entry with the maximum 'hit_score' in each group
idx = df_hmmer_RM.groupby('hit_id')['hit_score'].idxmax()
df_hmmer_RM_filtered = df_hmmer_RM.loc[idx]

#%% Filter defensefinder data for MTase systems only
df_hmmer_MTase = df_hmmer_RM_filtered[(df_hmmer_RM_filtered['gene_name'].str.contains("MTases")) | (df_hmmer_RM_filtered['gene_name'].str.contains("Type_IIG"))].copy()

#%% Determining whether the MTase is part of an RM system
df_systems_RM = df_systems[df_systems['type'] == "RM"]

# Defining a function to check whether a 'qseqid' is in any 'protein_in_syst' row
def check_RM_system(hit_id, protein_in_syst_column):
    for proteins in protein_in_syst_column:
        protein_list = proteins.split(',')
        if hit_id in protein_list:
            return True
    return False

# Checking whether MTase is part of any system
df_hmmer_MTase.loc[:, 'system'] = df_hmmer_MTase["hit_id"].apply(lambda x: check_RM_system(x, df_systems['protein_in_syst']))

# Checking wether MTase is part of an RM-system
df_hmmer_MTase.loc[:, 'RM_system'] = df_hmmer_MTase["hit_id"].apply(lambda x: check_RM_system(x, df_systems_RM['protein_in_syst']))

# Removing MTase hits which are part of another defense system 
df_hmmer_MTase_filtered = df_hmmer_MTase[~((df_hmmer_MTase['system'] == True) & (df_hmmer_MTase['RM_system'] == False))]
df_hmmer_MTase_filtered.drop('system', axis = 1, inplace = True) 

#%% Generating processed defensefinder.tsv
df_hmmer_MTase_filtered.to_csv(snakemake.output['DF_MTase_table'], sep = "\t", index = False)


#%% Generating MTase sequence files
MTase_sequence_file = snakemake.output['DF_MTase_AA']
prodigal_sequence_file = snakemake.input['prodigal_AAs']

# Load identifiers from the DataFrame
identifiers = df_hmmer_MTase_filtered['hit_id'].tolist()

with open(MTase_sequence_file, 'w') as output_file:
    for record in SeqIO.parse(prodigal_sequence_file, 'fasta'):
        # Check if the identifier is in the list
        if record.id.split(' ')[0] in identifiers:
            # Write the sequence to the new FASTA file
            SeqIO.write(record, output_file, 'fasta')



from nanomotif.constants import *
import nanomotif as nm
import pytest
from hypothesis import given, strategies as st
import itertools
import pandas as pd
import polars as pl
import polars.testing
import numpy as np


class TestMethylatedMotifOccourances:
    def test_known_motif_sequence(self):
        motif = nm.candidate.Motif('ACG', 0)
        sequence = 'TACGGACGCCACG'
        methylated_positions = np.array([1, 5])

        expected_methylated = np.array([1, 5])
        expected_nonmethylated = np.array([10])
        
        result = nm.evaluate.methylated_motif_occourances(motif, sequence, methylated_positions)
        np.testing.assert_array_equal(result[0], expected_methylated)
        np.testing.assert_array_equal(result[1], expected_nonmethylated)

class TestProcessSubpileup:
    def find_motifs(self):
        pileup_path = nm.datasets.geobacillus_plasmids_pileup_path()
        pileup = nm.dataload.load_pileup(pileup_path, min_fraction=0.5)
        modtype = "a"
        subpileup = pileup.pileup.filter(pl.col("mod_type") == modtype)

        assembly = nm.datasets.geobacillus_plasmids_assembly()
        contig = "contig_3"
        min_kl_divergence = 0.1
        padding = 20
        minimum_methylation_fraction_confident = 0.75

        result = nm.evaluate.process_subpileup(contig, modtype ,subpileup, assembly, min_kl_divergence, padding, minimum_methylation_fraction_confident)
        return result

    def test_no_errors(self):
        self.find_motifs()
    
    def test_correct_output(self):
        result = self.find_motifs()
        expected_output = pl.DataFrame({
            "sequence":[
                "...................GATC..................", 
                "................ACCCA....................", 
                "................CCAAAT...................", 
                "...............G[AG].GAAG[TC].................."],
            "score":[1.687542, 1.475938, 1.388421, 0.572978],
            "contig":["contig_3", "contig_3", "contig_3", "contig_3"],
            "mod_type":["a", "a", "a", "a"],
            "alpha":[706, 65, 80, 84],
            "beta":[3, 1, 1, 1]
        })
        polars.testing.assert_frame_equal(result, expected_output)


class TestProcessSampleParallel():
    def run_function(self):
        pileup_path = nm.datasets.geobacillus_plasmids_pileup_path()
        pileup = nm.dataload.load_pileup(pileup_path, min_fraction=0.5)
        pileup = pileup.pileup
        assembly = nm.datasets.geobacillus_plasmids_assembly()

        return nm.evaluate.process_sample_parallel(assembly, pileup, threads=2, seed=1)
    def test_no_errors(self):
        self.run_function()

    def test_reprocuible_result(self):
        res1 = self.run_function().drop("model").sort(["contig", "mod_type", "motif"])
        res2 = self.run_function().drop("model").sort(["contig", "mod_type", "motif"])
        polars.testing.assert_frame_equal(res1, res2)




import polars as pl
import pandas as pd
from nanomotif.binnary import data_processing as dp
from nanomotif.binnary import scoring as sc
from nanomotif.binnary.utils import split_bin_contig
import logging

def detect_contamination(motifs_scored_in_bins, bin_consensus, args):
    """
    Takes the bin_motif_binary and motifs_scored_in_bins DataFrames and performs the contamination detection analysis.
    Firstly bin_motif_binary is used to create a binary representation of the methylation status of each motif in each bin.
    This is the bases for the comparison. Only motifs that has a mean methylation value of at least 0.75 in bin_motif_binary are considered.
    
    Motifs_scored_in_bins is then used to create a binary representation of the methylation status of each motif in each contig.
    This is then compared to the binary representation of the bin to identify any mismatches.
    Only motifs that are observed more than 6 times are considered and must have a mean methylation value of at least 0.75-0.15.
    
    
    params:
        motifs_scored_in_bins: pd.DataFrame - DataFrame containing the motifs scored in each bin
        motifs_of_interest: list - List of motifs to be considered from bin_motif_binary
        args: argparse.Namespace - Namespace containing the arguments passed to the script
    """
    logger = logging.getLogger(__name__)
    logger.info("Starting contamination detection analysis...")
    motifs_scored_in_bins_wo_unbinned = motifs_scored_in_bins \
        .filter(~pl.col("bin_contig").str.contains("unbinned"))
    

    contig_bin_comparison_score, contigs_w_no_methylation = sc.compare_methylation_pattern_multiprocessed(
        motifs_scored_in_bins=motifs_scored_in_bins_wo_unbinned,
        bin_consensus=bin_consensus,
        mode="contamination",
        args=args,
        num_processes=args.threads
    )

    logger.info("Finding contamination in bins")
    
    contig_bin_comparison_score = split_bin_contig(contig_bin_comparison_score)
    
    # Filter contig_bin == bin and contig_bin_comparison_score > 0
    contamination_contigs = contig_bin_comparison_score \
        .filter(
            (pl.col("bin") == pl.col("contig_bin")) &
            (pl.col("binary_methylation_missmatch_score") > 0)
        )
    
    
    # contamination_contigs = contig_bin_comparison_score[
    #     # NOTE: This line also removes all contigs from bins with no methylation
    #     (contig_bin_comparison_score["bin"] == contig_bin_comparison_score["contig_bin"]) &
    #     (contig_bin_comparison_score["binary_methylation_missmatch_score"] > 0)
    # ]

    # logger.info("Finding alternative bin for contamination contigs")
    # # Find alternative bin for contamination contigs
    # ## Must have a perfect match
    # contamination_contigs_alternative_bin = contig_bin_comparison_score \
    #     .filter(
    #         (pl.col("bin") != pl.col("contig_bin")) &
    #         (pl.col("binary_methylation_missmatch_score") == 0) &
    #         (~pl.col("bin_compare").is_in(contigs_w_no_methylation))
    #     ) \
    #     .select(["contig", "bin", "binary_methylation_missmatch_score"]) \
    #     .rename(
    #         {
    #             "bin": "alternative_bin",
    #             "binary_methylation_missmatch_score": "alternative_bin_binary_methylation_missmatch_score"
    #         }
    #     )
    

    contamination_contigs = contamination_contigs \
        .drop("contig_bin") \
            .rename(
                {
                    "bin_compare": "bin_contig_compare"
                }
            ) \
        .sort("bin", "bin_contig_compare")

    
    logger.info("Contamination detection complete")
    
    return contamination_contigs



import polars as pl
import pandas as pd

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import gzip
import os
import logging

def load_data(args):
    """
    Load data for the analysis.
    """
    logger = logging.getLogger(__name__)
    logger.info("Loading data.")
    # Load the data from the provided files
    # Load the data from the provided files using Polars
    
    motifs_scored = pl.read_csv(args.motifs_scored, separator="\t")
    
    bin_motifs = pl.read_csv(args.bin_motifs, separator="\t")
    
    # Polars automatically infers headers; if args.contig_bins file doesn't have headers, you need to specify it
    contig_bins = pl.read_csv(args.contig_bins, separator="\t", has_header=False) \
        .rename({
            "column_1": "contig",
            "column_2": "bin"
        })

    logger.info("Data loaded.")
    return motifs_scored, bin_motifs, contig_bins

def read_fasta(file_path):
    # Check if the file exists
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")    
    
    # Check if the file has a valid FASTA extension
    valid_extensions = ['.fasta', '.fa', '.fna', '.gz']
    if not any(file_path.endswith(ext) for ext in valid_extensions):
        raise ValueError(f"Unsupported file extension. Please provide a FASTA file with one of the following extensions: {', '.join(valid_extensions)}")

    # Check if the file is a gzipped FASTA file
    if file_path.endswith('.gz'):
        with gzip.open(file_path, "rt") as handle:  # "rt" mode for reading as text
            return {record.id: str(record.seq) for record in SeqIO.parse(handle, "fasta")}
    else:
        # Read a regular (uncompressed) FASTA file
        with open(file_path, "r") as handle:
            return {record.id: str(record.seq) for record in SeqIO.parse(handle, "fasta")}


def write_bins_from_contigs(new_contig_bins, assembly_dict, output_dir):
    """
    Creates new bin files based on contig assignments in new_contig_bins DataFrame.

    Args:
    - new_contig_bins (DataFrame): DataFrame with contig and bin assignments.
    - assembly_dict (dict): Dictionary with contig IDs as keys and sequences as values, from the assembly file.
    - output_dir (str): Path to the directory where the new bin files will be saved.
    """
    logger = logging.getLogger(__name__)
    
    # check if the output directory exists
    logger.info(f"Writing bins to {output_dir}")
    output_bins_dir = os.path.join(output_dir)
    if not os.path.exists(output_bins_dir):
        os.makedirs(output_bins_dir)
    
    # Group the DataFrame by 'bin'
    grouped = new_contig_bins.groupby('bin')

    for bin_name, group in grouped:
        # Initialize a list to hold SeqRecord objects for the current bin
        bin_records = []

        for contig in group['contig']:
            # Create a SeqRecord object for each contig in the bin, if it exists in the assembly dictionary
            if contig in assembly_dict:
                seq_record = SeqRecord(Seq(assembly_dict[contig]), id=contig, description="")
                bin_records.append(seq_record)

        # Define the output file name for the current bin
        output_file = f"{bin_name}.fa"
        output_path = os.path.join(output_bins_dir, output_file)

        # Write the SeqRecord objects to a FASTA file
        with open(output_path, "w") as output_handle:
            SeqIO.write(bin_records, output_handle, "fasta")

        logger.info(f"Written {len(bin_records)} contigs to {output_file}")



def generate_output(output_df, outdir, filename, header=True):
    """
    Generate the output files for the analysis.
    """
    # If the output directory does not exist, create it
    if not os.path.exists(outdir) and outdir != "":
        os.makedirs(outdir)
    
    file_path = os.path.join(outdir, filename)
    
    # Generate the output files
    output_df.to_csv(file_path, sep="\t", index=False, header=header)


def prepare_bin_consensus(bin_motifs, args):
    """
    Prepares the bin_consensus_from_bin_motifs DataFrame by calculating the mean methylation per bin and motif_mod and converting it to binary.    
    """
    # Combine 'motif' and 'mod_type' into 'motif_mod'
    bin_motifs = bin_motifs.with_columns((pl.col("motif") + "_" + pl.col("mod_type") + "-" + pl.col("mod_position").cast(pl.Utf8)).alias("motif_mod"))

    # Calculate total motifs and mean methylation
    bin_motifs = bin_motifs.with_columns([
        (pl.col("n_mod_bin") + pl.col("n_nomod_bin")).alias("n_motifs"),
        (pl.col("n_mod_bin") / (pl.col("n_mod_bin") + pl.col("n_nomod_bin"))).alias("mean_methylation")
    ])

    # Create 'methylation_binary' column based on mean methylation cutoff
    bin_motifs = bin_motifs.with_columns((pl.col("mean_methylation") >= args.mean_methylation_cutoff).cast(int).alias("methylation_binary"))

    # Filter motifs that are not observed more than n_motif_bin_cutoff times
    bin_motifs = bin_motifs.filter(pl.col("n_motifs") >= args.n_motif_bin_cutoff)

    # Filter for rows where 'methylation_binary' is 1 and select relevant columns
    bin_motif_binary = bin_motifs.filter(pl.col("methylation_binary") == 1)[["bin", "motif_mod", "mean_methylation", "methylation_binary"]]

    # Create a binary index for bin and motif_mod
    # Assuming unique_bins and unique_motif_mods are lists of unique values
    unique_bins = bin_motif_binary["bin"].unique().to_list()
    unique_motif_mods = bin_motif_binary["motif_mod"].unique().to_list()

    # Create a DataFrame with all combinations of "bin" and "motif_mod"
    bin_motif_combinations = pl.DataFrame({
        "bin": [bin for bin in unique_bins for _ in unique_motif_mods],
        "motif_mod": [motif_mod for _ in unique_bins for motif_mod in unique_motif_mods]
    })

    # Join the combinations DataFrame with bin_motif_binary
    bin_motif_binary_enriched = bin_motif_combinations.join(
        bin_motif_binary, on=["bin", "motif_mod"], how="left"
    )

    # Fill missing values if necessary and adjust columns as needed
    bin_motif_binary_enriched = bin_motif_binary_enriched.with_columns([
        pl.col("mean_methylation").fill_null(0).alias("mean_methylation"),
        pl.col("methylation_binary").fill_null(0).alias("methylation_binary")
    ])
    

    return bin_motif_binary_enriched


def prepare_motifs_scored_in_bins(motifs_scored, motifs_of_interest, contig_bins):
    """
    Prepares the motifs_scored_in_bins DataFrame by merging with bin motifs, contig bins, and assembly stats,
    and calculates additional metrics like number of motifs and mean methylation per contig.
    """
    
    # Filter and enhance motifs_scored based on motifs_in_bins
    motifs_scored_in_bins = motifs_scored \
        .with_columns(
            (pl.col("motif") + "_" + pl.col("mod_type") + "-" + pl.col("mod_position").cast(pl.Utf8)).alias("motif_mod"),
            (pl.col("n_mod") + pl.col("n_nomod")).alias("n_motifs")
        ) \
        .filter(pl.col("motif_mod").is_in(motifs_of_interest)) \
        .with_columns((pl.col("n_mod") / pl.col("n_motifs")).alias("mean"))
    
    # Merge with contig_bins
    motifs_scored_in_bins = motifs_scored_in_bins.join(contig_bins, on="contig", how="left") \
        .with_columns(pl.col("bin").fill_null("unbinned")) \
        .with_columns((pl.col("bin") + "_" + pl.col("contig")).alias("bin_contig"))
    
    
    # Remove complement columns:
    # Identify columns that contain the word "complement"
    complement_columns = [col for col in motifs_scored_in_bins.columns if 'complement' in col]

    # Drop these columns from the DataFrame
    motifs_scored_in_bins = motifs_scored_in_bins.drop(complement_columns)

    
    return motifs_scored_in_bins


def remove_ambiguous_motifs_from_bin_consensus(motifs_scored_in_bins, args):
    # Remove motifs in bins where the majority of the mean methylation of motifs is in the range of 0.05-0.4
    contig_motif_mean_density = motifs_scored_in_bins \
        .filter(pl.col("bin") != "unbinned") \
        .filter(pl.col("n_motifs") >= args.n_motif_contig_cutoff) \
        .with_columns(
            ((pl.col("mean") > 0.05) & (pl.col("mean") < 0.4)).alias("is_ambiguous")
        )

    # Count the number of ambiguous motifs per bin
    # Group by 'bin' and 'motif_mod' and calculate the sum of 'is_ambiguous' and the total count in each group
    bin_consensus_ambiguous_motifs = contig_motif_mean_density.group_by(["bin", "motif_mod"]) \
        .agg(
            pl.col("is_ambiguous").sum().alias("total_ambiguous"),
            pl.col("is_ambiguous").count().alias("n_contigs_with_motif")
        ) \
        .with_columns((pl.col("total_ambiguous") / pl.col("n_contigs_with_motif")).alias("percentage_ambiguous")) \
        .filter(pl.col("percentage_ambiguous") <= args.ambiguous_motif_percentage_cutoff) \
        .select(["bin", "motif_mod"])
    
    
    return bin_consensus_ambiguous_motifs


# TODO: rename - calculate_bin_consensus_from_contigs
def construct_bin_consensus_from_motifs_scored_in_bins(motifs_scored_in_bins, args):
    """
    Constructs the bin_motifs_from_motifs_scored_in_bins DataFrame by filtering motifs that are not in bin_motif_binary,
    """
    
    # Remove motifs in bins where the majority of the mean methylation of motifs is in the range of 0.05-0.4
    bin_consensus_without_ambiguous_motifs = remove_ambiguous_motifs_from_bin_consensus(motifs_scored_in_bins, args)
    
    
    # Find n_motifs in bin TODO: rename bin_motifs_from_motifs_scored_in_bins to bin_consensus_from_motifs_scored_in_bins
    bin_consensus_from_motifs_scored_in_bins = motifs_scored_in_bins \
        .filter(pl.col("bin") != "unbinned") \
        .group_by(["bin", "motif_mod"]) \
        .agg(
            pl.col("n_mod").sum().alias("n_mod"),
            pl.col("n_nomod").sum().alias("n_nomod")
        ) \
        .with_columns((pl.col("n_mod") + pl.col("n_nomod")).alias("n_motifs_bin")) \
        .filter(pl.col("n_motifs_bin") > args.n_motif_bin_cutoff) \
        .with_columns((pl.col("n_mod") / pl.col("n_motifs_bin")).alias("mean_methylation"))
    
    bin_consensus_from_motifs_scored_in_bins_filtered = bin_consensus_from_motifs_scored_in_bins \
        .join(bin_consensus_without_ambiguous_motifs, on=["bin", "motif_mod"], how="inner")
    
    # Calculate standard deviation of methylation per bin and motif_mod
    bin_motifs_mean_and_sd = motifs_scored_in_bins \
        .filter(
            (pl.col("bin") != "unbinned") &  
            (pl.col("mean") > 0.1) &                                 # TODO: Remove this line if the negative cases should be used to determine methylation pattern.
            (pl.col("n_motifs") > args.n_motif_contig_cutoff)
        ) \
        .group_by(["bin", "motif_mod"]) \
        .agg(
            pl.col("mean").mean().alias("mean_methylation_bin"),
            pl.col("mean").std().fill_null(0.15/4).alias("std_methylation_bin"),
            pl.col("contig").count().alias("n_contigs")
        ) 
    
    # Merge with bin_motifs_from_motifs_scored_in_bins
    bin_consensus_from_motifs_scored_in_bins_filtered = bin_consensus_from_motifs_scored_in_bins_filtered.join(bin_motifs_mean_and_sd, on=["bin", "motif_mod"], how="left") \
        .with_columns((pl.col("mean_methylation") >= args.mean_methylation_cutoff).cast(int).alias("methylation_binary")) # Create a binary index for bin and motif_mod
    
    # bin_motifs_from_motifs_scored_in_bins = bin_motifs_from_motifs_scored_in_bins.merge(bin_motifs_mean_and_sd, on=["bin", "motif_mod"], how="left")
    
    # TODO: Does this make sense? All motifs with mean_methylation above 0.25 is called as methylated.

    
    return bin_consensus_from_motifs_scored_in_bins_filtered
    

# def calculate_binary_motif_comparison_matrix(motifs_scored_in_bins, args):
#     # Step 1 create bin_motif_from_motifs_scored_in_bins - basis for bin-contig comparison
#     bin_motifs_from_motifs_scored_in_bins = construct_bin_consensus_from_motifs_scored_in_bins(
#         motifs_scored_in_bins,
#         args
#     )
    
#     ## Filter motifs that are not observed more than n_motif_cutoff times
#     motifs_scored_in_contigs = motifs_scored_in_bins \
#         .filter(pl.col("n_motifs") > args.n_motif_contig_cutoff) \
#         .select(["bin_contig", "motif_mod", "mean"]) \
#         .rename({"bin_contig" : "bin_compare"})
    
#     # Merge bin_motifs_from_motifs_scored_in_bins and motifs_scored_in_contigs    
#     motif_binary_compare = bin_motifs_from_motifs_scored_in_bins.join(motifs_scored_in_contigs, on="motif_mod", how="left") 
        
#     # Calculate the mean methylation value for each motif in each bin
#     motif_binary_compare = motif_binary_compare.with_columns([
#         pl.when(pl.col("methylation_binary") == 1)
#         .then(
#             pl.when(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin") > 0.1)
#             .then(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin"))
#             .otherwise(0.1)
#         )
#         .otherwise(pl.lit(None))
#         .alias("methylation_mean_threshold")
#     ])

#     # Calculate the binary methylation value for each motif in each bin where the bin consensus is 1
#     motif_binary_compare = motif_binary_compare.with_columns([
#         pl.when((pl.col("methylation_binary") == 1) & 
#                 ((pl.col("mean") >= pl.col("methylation_mean_threshold")) | 
#                 (pl.col("mean") > 0.4)))
#         .then(1)
#         .when(pl.col("methylation_binary") == 1)
#         .then(0)
#         .otherwise(pl.lit(None))
#         .alias("methylation_binary_compare")
#     ])

#     # Calculate score for bin consensus is 0
#     motif_binary_compare = motif_binary_compare.with_columns([
#         pl.when(pl.col("methylation_binary") == 0)
#         .then(0.25)
#         .otherwise(pl.col("methylation_mean_threshold"))
#         .alias("methylation_mean_threshold"),

#         pl.when(pl.col("methylation_binary") == 0)
#         .then((pl.col("mean") >= 0.25).cast(pl.Int32))
#         .otherwise(pl.col("methylation_binary_compare"))
#         .alias("methylation_binary_compare")
#     ])
    
#     return motif_binary_compare



def load_contamination_file(contamination_file):
    """
    Load the contamination file from the provided path.
    """
    contamination = pl.read_csv(contamination_file, separator= "\t")
    
    # Check if the file contains the required columns
    # bin	bin_contig_compare	binary_methylation_missmatch_score	contig	alternative_bin	alternative_bin_binary_methylation_missmatch_score
    required_columns = ["bin", "bin_contig_compare", "binary_methylation_missmatch_score", "non_na_comparisons", "contig"]
    if not all(column in contamination.columns for column in required_columns):
        raise ValueError("The contamination file does not contain the required columns.")
    
    return contamination


def create_contig_bin_file(contig_bins, contamination, include=None):
    """
    Create a new contig_bin file based on the analysis results and contamination file.
    """
    # Remove contigs in the contamination file from the contig_bins
    contig_bins = contig_bins[~contig_bins["contig"].isin(contamination["contig"])]
    
    # Add the contigs in the include DataFrame to the contig_bins
    if include is not None:
        contig_bins = pd.concat([contig_bins, include[["contig", "bin"]]], ignore_index=True)
    
    # Sort the contig_bins by bin and contig
    contig_bins = contig_bins.sort_values(by=["bin", "contig"])
    
    return contig_bins


import pytest
from nanomotif.binnary import scoring as sc
from nanomotif.binnary import data_processing as dp
import polars as pl
from nanomotif.binnary.utils import split_bin_contig

from .conftest import MockArgs

def test_compare_methylation_pattern_contamination(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    
    """
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    args = MockArgs()
    
    motifs_scored_in_bins = motifs_scored_in_bins \
        .filter(~pl.col("bin_contig").str.contains("unbinned"))
    
    bin_consensus = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    # Define the corresponding choices for each condition
    # choices = [
    #     0,  # bin motif is methylated, contig motif is methylated
    #     1,  # bin motif is methylated, contig motif is not methylated
    #     1,  # bin motif is not methylated, contig motif is methylated
    #     0,  # bin motif is not methylated, contig motif is not methylated
    #     0,  # bin motif is methylated, contig motif is not observed
    #     0,  # bin motif is not methylated, contig motif is not observed
    # ]
    
    contig_bin_comparison_score, contigs_w_no_methylation = sc.compare_methylation_pattern_multiprocessed(
        motifs_scored_in_bins=motifs_scored_in_bins,
        bin_consensus=bin_consensus,
        mode="contamination",
        # choices=choices,
        args=args
    )
    contig_bin_comparison_score = split_bin_contig(contig_bin_comparison_score)
    print(contig_bin_comparison_score)
    print(contigs_w_no_methylation)
    
    contig_bin_comparison_score = contig_bin_comparison_score.to_pandas()
    
    assert contig_bin_comparison_score is not None
    assert set(contig_bin_comparison_score["bin"].unique()) == {'b3'}
    assert len(contig_bin_comparison_score["contig"].unique()) == 3
    # assert contig_bin_comparison_score[(contig_bin_comparison_score["bin"] == "b2") & (contig_bin_comparison_score["contig"] == "contig_4")]["binary_methylation_missmatch_score"].values[0] == 0
    

def test_compare_methylation_pattern_include(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    
    """
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    args = MockArgs()
    
    motifs_scored_in_bins = motifs_scored_in_bins \
        .filter(~pl.col("bin_contig").str.contains("unbinned"))
    
    bin_consensus = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    contig_bin_comparison_score, contigs_w_no_methylation = sc.compare_methylation_pattern_multiprocessed(
        motifs_scored_in_bins=motifs_scored_in_bins,
        bin_consensus=bin_consensus,
        mode="include",
        args=args
    )
    contig_bin_comparison_score = split_bin_contig(contig_bin_comparison_score)
    print(contig_bin_comparison_score)
    print(contigs_w_no_methylation)
    
    contig_bin_comparison_score = contig_bin_comparison_score.to_pandas()
    
    assert contig_bin_comparison_score is not None
    assert set(contig_bin_comparison_score["bin"].unique()) =={'b2', 'b3'} 
    assert len(contig_bin_comparison_score["contig"].unique()) == 2
    



from nanomotif.binnary import data_processing as dp
from nanomotif.binnary import utils as ut
import os
import polars as pl
from multiprocessing import Pool, Queue, Process, get_context
import logging
from logging.handlers import QueueHandler, QueueListener

# Set up logging
## Create queue for logging
# log_queue = Queue()

# ## Set up a listener to handle logs from the queue
# def setup_logging_queue(queue):
#     while True:
#         record = queue.get()
#         if record is None:  # Use None as a sentinel to stop the listener
#             break
#         logger = logging.getLogger(record.name)
#         logger.handle(record)

# def worker_setup_logging(queue):
#     q_handler = QueueHandler(queue)
#     logger = logging.getLogger()
#     logger.setLevel(logging.INFO)
#     logger.addHandler(q_handler)

def define_mean_methylation_thresholds(motif_binary_compare):
    """
    Define mean methylation thresholds for bin and contig motifs
    """
    # Calculate the mean methylation value for each motif in each bin
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when(pl.col("methylation_binary") == 1)
        .then(
            pl.when(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin") > 0.1)
            .then(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin"))
            .otherwise(0.1)
        )
        .otherwise(pl.lit(None))
        .alias("methylation_mean_threshold")
    ])

    # Calculate the binary methylation value for each motif in each bin where the bin consensus is 1
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when((pl.col("methylation_binary") == 1) & 
                ((pl.col("mean") >= pl.col("methylation_mean_threshold")) | 
                (pl.col("mean") > 0.4)))
        .then(1)
        .when(pl.col("methylation_binary") == 1)
        .then(0)
        .otherwise(pl.lit(None))
        .alias("methylation_binary_compare")
    ])

    # Calculate score for bin consensus is 0
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when(pl.col("methylation_binary") == 0)
        .then(0.25)
        .otherwise(pl.col("methylation_mean_threshold"))
        .alias("methylation_mean_threshold"),

        pl.when(pl.col("methylation_binary") == 0)
        .then((pl.col("mean") >= 0.25).cast(pl.Int32))
        .otherwise(pl.col("methylation_binary_compare"))
        .alias("methylation_binary_compare")
    ])
    
    return motif_binary_compare



def compare_methylation_pattern(motif_binary_compare):
    """
    Compares the methylation pattern between bin and contig motifs using Polars and calculates the motif_comparison_score.
    """
    motif_comparison_score = (
        pl.when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 1))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 0))
        .then(1)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 1))
        .then(1)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 0))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'].is_null()))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'].is_null()))
        .then(0)
        .otherwise(pl.lit(None))
    )

    # Add the 'motif_comparison_score' column to the DataFrame
    motif_binary_compare = motif_binary_compare.with_columns(motif_comparison_score.alias("motif_comparison_score"))
    
    # Group by bin and bin_compare and calculate the sum of the motif_comparison_score and the count of non-NA values
    contig_bin_comparison_score = motif_binary_compare \
        .group_by(["bin", "bin_compare"]) \
        .agg([
            pl.sum("motif_comparison_score").alias("binary_methylation_missmatch_score"),
            pl.count("motif_comparison_score").alias("non_na_comparisons")
        ])
        
    return contig_bin_comparison_score


def process_bin_contig(
    bin_contig, 
    bin_motifs_from_motifs_scored_in_bins, 
    motifs_scored_in_contigs, 
    mode, 
    args
):
    """
    This function processes a single contig and compares the methylation pattern between the contig and the bin.
    Depending on the mode, the function will either look for contamination or include contigs.
    
    Parameters:
    - bin_contig: str
        The contig to process
    - bin_motifs_from_motifs_scored_in_bins: polars.DataFrame
        The motifs scored in the bin
    - motifs_scored_in_contigs: polars.DataFrame
        The motifs scored in the contigs
    - mode: str
        The mode to run the comparison in. Either "contamination" or "include"
    - args: argparse.Namespace
        The arguments passed to the script
    
    Returns:
    - contig_bin_comparison_score: polars.DataFrame
        The comparison score for the contig
    """
    # worker_setup_logging(log_queue)
    # logger = logging.getLogger(__name__)
    # logger.info(f"Processing {bin_contig}")
    try:
        bin = ut.get_bin(bin_contig)
        
        if mode == "contamination":
            motif_binary_compare = bin_motifs_from_motifs_scored_in_bins \
                .filter(pl.col("bin") == bin) \
                .join(
                    motifs_scored_in_contigs.filter(pl.col("bin_compare") == bin_contig),
                    on="motif_mod"
                )
        if mode == "include":
            motif_binary_compare = bin_motifs_from_motifs_scored_in_bins \
                .filter(pl.col("bin") != bin) \
                .join(
                    motifs_scored_in_contigs.filter(pl.col("bin_compare") == bin_contig),
                    on="motif_mod"
                )
        
        # Define methylation thresholds
        motif_binary_compare = define_mean_methylation_thresholds(motif_binary_compare)

        if args.save_scores:
            path = os.path.join(args.out, "scores", args.command, "binary_compare" + bin_contig + ".csv")
            motif_binary_compare.write_csv(path)

        # Calculate the comparison score regardless of methylation presence
        contig_bin_comparison_score = compare_methylation_pattern(motif_binary_compare)
        
        # Check if the contig has no methylation and note it, but do not exclude it from further processing
        contigHasNMethylation = motif_binary_compare.filter(pl.col("methylation_binary_compare") == 1).height
        # logger.info(f"Finished processing {bin_contig}. Contig has {contigHasNMethylation} positive methylation comparisons.")
        
        if mode == "include":
            contig_bin_comparison_score = contig_bin_comparison_score \
                .filter(
                    pl.col("binary_methylation_missmatch_score") == 0
                )
        if mode == "contamination":
            contig_bin_comparison_score = contig_bin_comparison_score \
                .filter(
                    pl.col("binary_methylation_missmatch_score") > 0
                )
                
            
        if contigHasNMethylation == 0:
            return contig_bin_comparison_score, bin_contig
        
        return contig_bin_comparison_score, None
    except Exception as e:
        error_log_path = os.path.join(args.out, "logs", f"{args.command}_{bin_contig}_error.err")
        
        with open(error_log_path, 'w') as error_file:
            error_file.write(f"Error processing {bin_contig}: {str(e)}\n")
        return None, None

def compare_methylation_pattern_multiprocessed(motifs_scored_in_bins, bin_consensus, mode, args, num_processes=1):
    logger = logging.getLogger(__name__)
    logger.info("Starting comparison of methylation patterns")
    
    motifs_scored_in_contigs = motifs_scored_in_bins \
        .filter(pl.col("n_motifs") >= args.n_motif_contig_cutoff) \
        .select(["bin_contig", "motif_mod", "mean"]) \
        .rename({"bin_contig": "bin_compare"})
    
    if args.save_scores:
        dir = os.path.join(args.out, "scores", args.command)
        if not os.path.exists(dir):
            os.makedirs(dir)

    comparison_score = pl.DataFrame()
    contigs_w_no_methylation = []

    with get_context("spawn").Pool(processes=num_processes) as pool:
        results = pool.starmap(
            process_bin_contig,
            [
                (bin_contig, bin_consensus, motifs_scored_in_contigs, mode, args)
                for bin_contig in motifs_scored_in_contigs.select("bin_compare").unique()["bin_compare"]
            ]
        )
        
    for result, no_methylation in results:
        if result is not None:
            comparison_score = pl.concat([comparison_score, result])
        if no_methylation is not None:
            contigs_w_no_methylation.append(no_methylation)
        
    return comparison_score, contigs_w_no_methylation

